// NOTE - this is generated by unicode_ranges.pl
module.exports = {
  // these patterns are simply XID_Start followed by XID_Continue, minus things which don't match perl's \p{Word}
  identifier: /[_\p{XID_Start}][\p{XID_Continue}]*/v,
  // this adds in any amount of :: before or in between identifiers
  bareword:   /((::)|([_\p{XID_Start}][\p{XID_Continue}]*))+/v,
}
